@@ -1,6 +1,6 @@
 /* -*- Mode: C; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
 /*
- *  Copyright (C) 2007-2010  Kouhei Sutou <kou@clear-code.com>
+ *  Copyright (C) 2007-2011  Kouhei Sutou <kou@clear-code.com>
  *
  *  This library is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU Lesser General Public License as published by
@@ -94,6 +94,8 @@ struct _CutTestContextPrivate
     gchar *actual;
     CutTestResult *current_result;
     guint user_message_jump_nest;
+    GThread *main_thread;
+    CutTestContext *parent;
 };
 
 enum
@@ -103,7 +105,8 @@ enum
     PROP_TEST_SUITE,
     PROP_TEST_CASE,
     PROP_TEST_ITERATOR,
-    PROP_TEST
+    PROP_TEST,
+    PROP_PARENT
 };
 
 G_DEFINE_TYPE (CutTestContext, cut_test_context, G_TYPE_OBJECT)
@@ -228,9 +231,23 @@ cut_test_context_class_init (CutTestContextClass *klass)
                                G_PARAM_READWRITE);
     g_object_class_install_property(gobject_class, PROP_TEST, spec);
 
+    spec = g_param_spec_object("parent",
+                               "Parent",
+                               "The parent test context of the test context",
+                               CUT_TYPE_TEST_CONTEXT,
+                               G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY);
+    g_object_class_install_property(gobject_class, PROP_PARENT, spec);
+
     g_type_class_add_private(gobject_class, sizeof(CutTestContextPrivate));
 }
 
+static void
+fixture_data_free (gpointer data)
+{
+    GString *string = data;
+    g_string_free(string, TRUE);
+}
+
 static void
 cut_test_context_init (CutTestContext *context)
 {
@@ -255,7 +272,7 @@ cut_test_context_init (CutTestContext *context)
 
     priv->fixture_data_dir = NULL;
     priv->cached_fixture_data = g_hash_table_new_full(g_str_hash, g_str_equal,
-                                                      g_free, g_free);
+                                                      g_free, fixture_data_free);
 
     priv->backtrace = NULL;
 
@@ -270,6 +287,10 @@ cut_test_context_init (CutTestContext *context)
     priv->current_result = NULL;
 
     priv->user_message_jump_nest = 0;
+
+    priv->main_thread = g_thread_self();
+
+    priv->parent = NULL;
 }
 
 static void
@@ -390,9 +411,33 @@ dispose (GObject *object)
     clear_additional_test_result_data(priv);
     clear_current_result(priv);
 
+    if (priv->parent) {
+        g_object_unref(priv->parent);
+        priv->parent = NULL;
+    }
+
     G_OBJECT_CLASS(cut_test_context_parent_class)->dispose(object);
 }
 
+static void
+set_parent (CutTestContext *context, CutTestContext *parent)
+{
+    CutTestContextPrivate *priv;
+
+    if (!parent)
+        return;
+
+    priv = CUT_TEST_CONTEXT_GET_PRIVATE(context);
+    priv->parent = g_object_ref(parent);
+    g_object_set(G_OBJECT(context),
+                 "run-context", cut_test_context_get_run_context(parent),
+                 "test-suite", cut_test_context_get_test_suite(parent),
+                 "test-case", cut_test_context_get_test_case(parent),
+                 "test-iterator", cut_test_context_get_test_iterator(parent),
+                 "test", cut_test_context_get_test(parent),
+                 NULL);
+}
+
 static void
 set_property (GObject      *object,
               guint         prop_id,
@@ -402,22 +447,25 @@ set_property (GObject      *object,
     CutTestContext *context = CUT_TEST_CONTEXT(object);
 
     switch (prop_id) {
-      case PROP_RUN_CONTEXT:
+    case PROP_RUN_CONTEXT:
         cut_test_context_set_run_context(context, g_value_get_object(value));
         break;
-      case PROP_TEST_SUITE:
+    case PROP_TEST_SUITE:
         cut_test_context_set_test_suite(context, g_value_get_object(value));
         break;
-      case PROP_TEST_CASE:
+    case PROP_TEST_CASE:
         cut_test_context_set_test_case(context, g_value_get_object(value));
         break;
-      case PROP_TEST_ITERATOR:
+    case PROP_TEST_ITERATOR:
         cut_test_context_set_test_iterator(context, g_value_get_object(value));
         break;
-      case PROP_TEST:
+    case PROP_TEST:
         cut_test_context_set_test(context, g_value_get_object(value));
         break;
-      default:
+    case PROP_PARENT:
+        set_parent(context, g_value_get_object(value));
+        break;
+    default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
         break;
     }
@@ -482,6 +530,12 @@ void
 cut_test_context_current_push (CutTestContext *context)
 {
     GPtrArray *contexts;
+    CutTestContextPrivate *priv;
+
+    priv = CUT_TEST_CONTEXT_GET_PRIVATE(context);
+    if (priv->main_thread != g_thread_self()) {
+        context = cut_test_context_new_sub(context);
+    }
 
     contexts = g_static_private_get(&current_context_private);
     if (!contexts) {
@@ -538,6 +592,14 @@ cut_test_context_new_empty (void)
     return cut_test_context_new(NULL, NULL, NULL, NULL, NULL);
 }
 
+CutTestContext *
+cut_test_context_new_sub (CutTestContext *parent)
+{
+    return g_object_new(CUT_TYPE_TEST_CONTEXT,
+                        "parent", parent,
+                        NULL);
+}
+
 CutRunContext *
 cut_test_context_get_run_context (CutTestContext *context)
 {
@@ -776,9 +838,28 @@ cut_test_context_get_jump (CutTestContext *context)
 void
 cut_test_context_long_jump (CutTestContext *context)
 {
+    gboolean same_thread_p = TRUE;
     CutTestContextPrivate *priv = CUT_TEST_CONTEXT_GET_PRIVATE(context);
 
-    longjmp(*(priv->jump_buffer), 1);
+    if (priv->parent) {
+        CutTestContextPrivate *parent_priv;
+        parent_priv = CUT_TEST_CONTEXT_GET_PRIVATE(priv->parent);
+        if (parent_priv->main_thread != g_thread_self()) {
+            same_thread_p = FALSE;
+        }
+    } else {
+        if (priv->main_thread != g_thread_self()) {
+            same_thread_p = FALSE;
+        }
+    }
+
+    if (cut_test_context_in_user_message_jump(context) || same_thread_p) {
+        longjmp(*(priv->jump_buffer), 1);
+    } else {
+        cut_test_context_current_pop();
+        g_object_unref(context);
+        g_thread_exit(NULL);
+    }
 }
 
 void
@@ -1097,6 +1178,21 @@ cut_test_context_take_g_hash_table (CutTestContext *context,
                                                (CutDestroyFunction)g_hash_table_unref);
 }
 
+static void
+cut_test_context_g_string_free (GString *string)
+{
+    g_string_free(string, TRUE);
+}
+
+GString *
+cut_test_context_take_g_string (CutTestContext *context,
+                                GString        *string)
+{
+    return (GString *)cut_test_context_take(context,
+                                            string,
+                                            (CutDestroyFunction) cut_test_context_g_string_free);
+}
+
 int
 cut_test_context_trap_fork (CutTestContext *context)
 {
@@ -1268,106 +1364,106 @@ cut_test_context_set_fixture_data_dir (CutTestContext *context,
     va_end(args);
 }
 
-gchar *
-cut_test_context_build_fixture_data_path_va_list (CutTestContext *context,
-                                                  const gchar *path,
-                                                  va_list args)
+const gchar *
+cut_test_context_build_fixture_path_va_list (CutTestContext *context,
+                                             const gchar *path,
+                                             va_list args)
 {
     CutTestContextPrivate *priv;
-    gchar *concatenated_path, *full_path;
+    gchar *concatenated_path;
+    const gchar *full_path;
 
     priv = CUT_TEST_CONTEXT_GET_PRIVATE(context);
 
     concatenated_path = cut_utils_build_path_va_list(path, args);
     if (g_path_is_absolute(concatenated_path)) {
-        full_path = concatenated_path;
+        full_path = cut_test_context_take_string(context, concatenated_path);
     } else {
+        gchar *resolved_path;
         if (priv->fixture_data_dir) {
-            full_path = g_build_filename(priv->fixture_data_dir,
-                                         concatenated_path,
-                                         NULL);
+            resolved_path = g_build_filename(priv->fixture_data_dir,
+                                             concatenated_path,
+                                             NULL);
         } else {
-            full_path = cut_utils_expand_path(concatenated_path);
+            resolved_path = cut_utils_expand_path(concatenated_path);
         }
-        g_free(concatenated_path);
+        full_path = cut_test_context_take_string(context, resolved_path);
     }
-
-    return full_path;
-}
-
-gchar *
-cut_test_context_build_fixture_data_path (CutTestContext *context,
-                                          const gchar *path, ...)
-{
-    va_list args;
-    gchar *full_path;
-
-    va_start(args, path);
-    full_path = cut_test_context_build_fixture_data_path_va_list(context,
-                                                                 path, args);
-    va_end(args);
+    g_free(concatenated_path);
 
     return full_path;
 }
 
 const gchar *
-cut_test_context_get_fixture_data_string_va_list (CutTestContext *context,
-                                                  GError **error,
-                                                  gchar **full_path,
-                                                  const gchar *path,
-                                                  va_list args)
+cut_test_context_build_fixture_path (CutTestContext *context,
+                                     const gchar *path, ...)
+{
+    va_list args;
+    const gchar *full_path;
+
+    va_start(args, path);
+    full_path = cut_test_context_build_fixture_path_va_list(context, path, args);
+    va_end(args);
+
+    return full_path;
+}
+
+GString *
+cut_test_context_get_fixture_data_va_list (CutTestContext *context,
+                                           GError **error,
+                                           const gchar **full_path,
+                                           const gchar *path,
+                                           va_list args)
 {
     CutTestContextPrivate *priv;
-    gpointer value;
-    gchar *_full_path;
+    GString *data;
+    const gchar *fixture_full_path;
 
     if (!path)
         return NULL;
 
     priv = CUT_TEST_CONTEXT_GET_PRIVATE(context);
 
-    _full_path = cut_test_context_build_fixture_data_path_va_list(context,
-                                                                  path, args);
-    value = g_hash_table_lookup(priv->cached_fixture_data, _full_path);
-    if (!value) {
+    fixture_full_path = cut_test_context_build_fixture_path_va_list(context,
+                                                                    path, args);
+    data = g_hash_table_lookup(priv->cached_fixture_data, fixture_full_path);
+    if (!data) {
         gchar *contents;
         gsize length;
 
-        if (g_file_get_contents(_full_path, &contents, &length, error)) {
+        if (g_file_get_contents(fixture_full_path, &contents, &length, error)) {
+            data = g_string_new_len(contents, length);
+            g_free(contents);
             g_hash_table_insert(priv->cached_fixture_data,
-                                g_strdup(_full_path),
-                                contents);
-            value = contents;
+                                g_strdup(fixture_full_path),
+                                data);
         }
     }
 
     if (full_path)
-        *full_path = _full_path;
-    else
-        g_free(_full_path);
+        *full_path = fixture_full_path;
 
-    return value;
+    return data;
 }
 
-const gchar *
-cut_test_context_get_fixture_data_string (CutTestContext *context,
-                                          GError **error,
-                                          gchar **full_path,
-                                          const gchar *path, ...)
+GString *
+cut_test_context_get_fixture_data (CutTestContext *context,
+                                   GError **error,
+                                   const gchar **full_path,
+                                   const gchar *path, ...)
 {
-    const gchar *value;
+    GString *data;
     va_list args;
 
     if (!path)
         return NULL;
 
     va_start(args, path);
-    value = cut_test_context_get_fixture_data_string_va_list(context, error,
-                                                             full_path,
-                                                             path, args);
+    data = cut_test_context_get_fixture_data_va_list(context, error,
+                                                     full_path, path, args);
     va_end(args);
 
-    return value;
+    return data;
 }
 
 gchar *
